import ctypes

from glob import glob
from typing import Final, Callable

from edge import eventlog
from edge.services.types.cam import CamServiceConf, ICamService, SpeedcamConf
from edge.support.speed import FastVehicleError, SlowVehicleError, map_speed_int
from edge.types.exceptions import InternalError, ParamsError
from edge.types.track import LrTrackItem


class CQuad(ctypes.Structure):
    _fields_ = [
        ('x0', ctypes.c_int), ('y0', ctypes.c_int),
        ('x1', ctypes.c_int), ('y1', ctypes.c_int),
        ('x2', ctypes.c_int), ('y2', ctypes.c_int),
        ('x3', ctypes.c_int), ('y3', ctypes.c_int)
    ]

class CPoint(ctypes.Structure):
    _fields_ = [
        ('x', ctypes.c_int), ('y', ctypes.c_int)
    ]


kLLM_CAM_STAT__DONE = 0
kLLM_CAM_STAT__FAIL_PARAMS = 1
kLLM_CAM_STAT__FAIL_SLOW_VEHICLE = 2
kLLM_CAM_STAT__FAIL_FAST_VEHICLE = 3


class _Api:
    _lib = ctypes.CDLL(glob('/opt/edge/lib/liblandmark*.so')[0])

    # bool
    # LlmCamCreate3(const char *aCamIntrinsicConf, const char *aCamExtrinsicConf, llm_cam_t *aSelf)
    camCreate3 = _lib.LlmCamCreate3
    camCreate3.argtypes = (
        ctypes.c_char_p, ctypes.c_char_p,
        ctypes.POINTER(ctypes.c_void_p))

    camCreate3.restype = ctypes.c_bool

    # void LlmCamDestroy(void* aSelf);
    camDestroy = _lib.LlmCamDestroy
    camDestroy.argtypes = (ctypes.c_void_p,)

    # bool
    # LlmCamMkImagePoint(
    #     void *aSelf,
    #     double aSceneL,
    #     double aSceneU,
    #     double aSceneF,
    #     double *anImageL,
    #     double *anImageT);
    mkImagePoint = _lib.LlmCamMkImagePoint
    mkImagePoint.argtypes = (
        ctypes.c_void_p, ctypes.c_double, ctypes.c_double, ctypes.c_double,
        ctypes.POINTER(ctypes.c_double),
        ctypes.POINTER(ctypes.c_double))
    mkImagePoint.restype = ctypes.c_bool

    # bool
    # LlmCamMkScenePointU(void *aSelf,
    #     double aFrameL, double aFrameT, double aSceneU,
    #     double *aSceneL, double *aSceneF);
    mkScenePoint = _lib.LlmCamMkScenePointU
    mkScenePoint.argtypes = (
        ctypes.c_void_p, ctypes.c_double, ctypes.c_double, ctypes.c_double,
        ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double))

    mkScenePoint.restype = ctypes.c_bool

    # int
    # LlmCamAssertPointSpeed(llm_cam_t aSelf,
    #     const char *aPlateText, uint64_t aPointTs1, uint64_t aPointTs2,
    #     struct llm_point *aPoint1, struct llm_point *aPoint2, double aPointU,
    #     bool aSlowVehicle, bool aMainDirU,
    #     double aMainSpeed, double *aControlSpeed);
    assertPointSpeed = _lib.LlmCamAssertPointSpeed
    assertPointSpeed.argtypes = (
        ctypes.c_void_p,
        ctypes.c_char_p, ctypes.c_uint64, ctypes.c_uint64,
        ctypes.POINTER(CPoint), ctypes.POINTER(CPoint), ctypes.c_double,
        ctypes.c_bool, ctypes.c_bool,
        ctypes.c_double, ctypes.POINTER(ctypes.c_double))

    assertPointSpeed.restype = ctypes.c_int

    # int
    # LlmCamAssertPlateSpeed(llm_cam_t aSelf,
    #     const char *aPlateText, uint64_t aPlateTs1, uint64_t aPlateTs2,
    #     struct llm_quad *aPlateQ1, struct llm_quad *aPlateQ2, double *aPlateU,
    #     bool aSlowVehicle, bool aMainDirU,
    #     double aMainSpeed, double *aControlSpeed);
    assertPlateSpeed = _lib.LlmCamAssertPlateSpeed
    assertPlateSpeed.argtypes = (
        ctypes.c_void_p,
        ctypes.c_char_p, ctypes.c_uint64, ctypes.c_uint64,
        ctypes.POINTER(CQuad), ctypes.POINTER(CQuad), ctypes.POINTER(ctypes.c_double),
        ctypes.c_bool, ctypes.c_bool,
        ctypes.c_double, ctypes.POINTER(ctypes.c_double))

    assertPlateSpeed.restype = ctypes.c_int


class _SpeedcamStats:
    def __init__(self):
        self.cam_speed_total: int = 0
        self.cam_speed_absent: int = 0
        self.cam_speed_invalid_car: int = 0
        self.cam_speed_invalid_bus: int = 0


class _CamService(ICamService):
    def __init__(self, conf: CamServiceConf):
        super().__init__()

        self._handle = ctypes.c_void_p(0)
        answer = _Api.camCreate3(
            bytes(conf.intrinsics_path, 'ascii'),
            bytes(conf.extrinsics_path, 'ascii'),
            self._handle)

        if not answer:
            eventlog.log_e('fail: camCreate3 (camera:<{}>, location:<{}>)'.format(
                conf.intrinsics_path, conf.extrinsics_path))
            raise InternalError

        self._speedcam2_conf: Final[SpeedcamConf | None] = conf.speedcam2
        self._speedcam2_stats = _SpeedcamStats()

    def destroy(self) -> None:
        _Api.camDestroy(self._handle)

    def mk_image_point(self, scene_l: float, scene_u: float, scene_f: float) -> tuple[int, int]:
        image_l = ctypes.c_double()
        image_t = ctypes.c_double()
        success: bool = _Api.mkImagePoint(self._handle, scene_l, scene_u, scene_f, image_l, image_t)
        if not success:
            eventlog.log_i('fail: mkImagePoint')
            raise ParamsError
        return int(image_l.value), int(image_t.value)

    def mk_world_point(self, image_x: int, image_y: int, world_t: float) -> tuple[float, float]:
        scene_l = ctypes.c_double()
        scene_f = ctypes.c_double()
        success: bool = _Api.mkScenePoint(self._handle, image_x, image_y, world_t, scene_l, scene_f)
        if not success:
            eventlog.log_i('fail: mkScenePoint')
            raise ParamsError

        return scene_f.value, scene_l.value

    def _assert_cam_speed_vr(self,
                             vrp_text_utf8: bytes,
                             main_speed_kmh: int,
                             slow_vehicle: bool,
                             min_item_with_vr: LrTrackItem,
                             max_item_with_vr: LrTrackItem,
                             main_dir_u: bool = False) -> int:

        self._speedcam2_stats.cam_speed_total += 1

        c_control_speed_kmh = ctypes.c_double()

        c_plate_text = None
        if self._speedcam2_conf.with_log:
            c_plate_text = vrp_text_utf8

        cpoint1 = CPoint(int((min_item_with_vr.vehicle_rect.il + min_item_with_vr.vehicle_rect.ir) / 2),
                         min_item_with_vr.vehicle_rect.ib)

        cpoint2 = CPoint(int((max_item_with_vr.vehicle_rect.il + max_item_with_vr.vehicle_rect.ir) / 2),
                         max_item_with_vr.vehicle_rect.ib)

        c_plate_up = ctypes.c_double(0)
        c_answer = _Api.assertPointSpeed(
            self._handle, c_plate_text, min_item_with_vr.ts, max_item_with_vr.ts,
            ctypes.byref(cpoint1), ctypes.byref(cpoint2), c_plate_up,
            slow_vehicle, main_dir_u,
            main_speed_kmh, ctypes.byref(c_control_speed_kmh))

        control_speed_kmh: float = c_control_speed_kmh.value
        if c_answer == kLLM_CAM_STAT__FAIL_SLOW_VEHICLE:
            eventlog.log_i(
                'fail: speedcam (plate:{}, reason:slow-vehicle, main-speed:{:3f})'.format(
                    vrp_text_utf8, main_speed_kmh))
            self._speedcam2_stats.cam_speed_invalid_bus += 1
            raise SlowVehicleError()

        elif c_answer == kLLM_CAM_STAT__FAIL_FAST_VEHICLE:
            eventlog.log_i(
                'fail: speedcam (plate:{}, reason:fast-vehicle, main-speed:{:3f})'.format(
                    vrp_text_utf8, main_speed_kmh))
            self._speedcam2_stats.cam_speed_invalid_car += 1
            raise FastVehicleError()

        elif c_answer == kLLM_CAM_STAT__FAIL_PARAMS:
            eventlog.log_i(f'fail: speedcam (plate:{vrp_text_utf8}, reason:invalid-params)')
            self._speedcam2_stats.cam_speed_absent += 1
            raise ParamsError()

        return map_speed_int(control_speed_kmh)

    def _assert_cam_speed_vrp(self,
                              vrp_text_utf8: bytes,
                              main_speed_kmh: int,
                              slow_vehicle: bool,
                              min_lr_item: LrTrackItem,
                              max_lr_item: LrTrackItem,
                              main_dir_u: bool = False) -> int:

        if self._speedcam2_conf is None:
            raise NotImplementedError

        self._speedcam2_stats.cam_speed_total += 1

        if min_lr_item.ts == max_lr_item.ts:
            eventlog.log_i('fail')
            raise ParamsError

        c_double_p = ctypes.POINTER(ctypes.c_double)
        c_plate_up_p: c_double_p | None = None
        if self._speedcam2_conf.avg_plate_u is not None:
            c_plate_up = ctypes.c_double(self._speedcam2_conf.avg_plate_u)
            c_plate_up_p = c_double_p(c_plate_up)

        c_control_speed_kmh = ctypes.c_double()

        c_plate_text = None
        if self._speedcam2_conf.with_log:
            c_plate_text = vrp_text_utf8

        quad1 = min_lr_item.plate_quad
        quad2 = max_lr_item.plate_quad

        cquad1 = CQuad(quad1.x0, quad1.y0, quad1.x1, quad1.y1, quad1.x2, quad1.y2, quad1.x3, quad1.y3)
        cquad2 = CQuad(quad2.x0, quad2.y0, quad2.x1, quad2.y1, quad2.x2, quad2.y2, quad2.x3, quad2.y3)

        c_answer = _Api.assertPlateSpeed(
            self._handle, c_plate_text, min_lr_item.ts, max_lr_item.ts,
            ctypes.byref(cquad1), ctypes.byref(cquad2), c_plate_up_p,
            slow_vehicle, main_dir_u,
            main_speed_kmh, ctypes.byref(c_control_speed_kmh))

        control_speed_kmh: float = c_control_speed_kmh.value
        if c_answer == kLLM_CAM_STAT__FAIL_SLOW_VEHICLE:
            eventlog.log_i(
                'fail: speedcam (plate:{}, reason:slow-vehicle, main-speed:{:3f})'.format(
                    vrp_text_utf8, main_speed_kmh))
            self._speedcam2_stats.cam_speed_invalid_bus += 1
            raise SlowVehicleError()

        elif c_answer == kLLM_CAM_STAT__FAIL_FAST_VEHICLE:
            eventlog.log_i(
                'fail: speedcam (plate:{}, reason:fast-vehicle, main-speed:{:3f})'.format(
                    vrp_text_utf8, main_speed_kmh))
            self._speedcam2_stats.cam_speed_invalid_car += 1
            raise FastVehicleError()

        elif c_answer == kLLM_CAM_STAT__FAIL_PARAMS:
            eventlog.log_i(f'fail: speedcam (plate:{vrp_text_utf8}, reason:invalid-params)')
            self._speedcam2_stats.cam_speed_absent += 1
            raise ParamsError()

        return map_speed_int(control_speed_kmh)

    def assert_cam_speed_2(self,
                           vrp_text_utf8: bytes,
                           main_speed_kmh: int,
                           lr_items: list[LrTrackItem],
                           mmr_class: str,
                           main_dir_u: bool = False) -> int:

        min_item_with_vr = None
        for item in lr_items:
            if item.vehicle_rect is not None:
                min_item_with_vr = item
                break

        max_item_with_vr = None
        for item in reversed(lr_items):
            if item.vehicle_rect is not None:
                max_item_with_vr = item
                break

        prefer_vr = False
        if None not in [min_item_with_vr, max_item_with_vr]:
            if min_item_with_vr.ts != max_item_with_vr.ts:
                prefer_vr = True

        slow_vehicle = mmr_class in ['UNKNOWN', 'BUS', 'LT', 'HT']
        if prefer_vr:
            # vehicle rect
            return self._assert_cam_speed_vr(
                vrp_text_utf8, main_speed_kmh, slow_vehicle,
                min_item_with_vr, max_item_with_vr, main_dir_u)
        else:
            # vehicle registration plate
            return self._assert_cam_speed_vrp(
                vrp_text_utf8, main_speed_kmh, slow_vehicle,
                lr_items[0], lr_items[-1], main_dir_u)

    def collect_stats(self, json_handler: Callable[[dict], None]) -> None:
        cam_total_fail = (self._speedcam2_stats.cam_speed_absent +
                          self._speedcam2_stats.cam_speed_invalid_bus +
                          self._speedcam2_stats.cam_speed_invalid_car)

        json_handler({
            'type': 'cam',
            'name': 'cam.0',
            'stat': {
                'assert-cam-speed': {
                    'count': [
                        self._speedcam2_stats.cam_speed_total,
                        self._speedcam2_stats.cam_speed_total - cam_total_fail,
                        cam_total_fail
                    ],
                    'speed-absent': self._speedcam2_stats.cam_speed_absent,
                    'speed-invalid-bus': self._speedcam2_stats.cam_speed_invalid_bus,
                    'speed-invalid-car': self._speedcam2_stats.cam_speed_invalid_car
                }
            }
        })


def init_service(conf: CamServiceConf) -> ICamService:
    return _CamService(conf)
