from typing import Dict, List

from edge import eventlog
from edge.types.exceptions import JnodeError
from renderer.services.types.jnode.tl import decode__tl_conf
from renderer.services.types.renderer import LrAnnoConf, RaAnnoConf, RendererConf, MarkingsConf, MarkingLine2Conf, RefPoint2d, MarkingLine3Conf, RefPoint3d
from renderer.services.types.tl import TlConf


def decode__lr_anno_conf(jnode: Dict) -> LrAnnoConf:
    """
    :param jnode:
        main.conf#/Renderer/profile/vehicle/lr

    :raise JnodeError:
    """
    modes = [LrAnnoConf.MODE_NEVER, LrAnnoConf.MODE_ALWAYS]

    # unpack #/plate
    plate_mode = jnode.get('plate', LrAnnoConf.MODE_NEVER)
    if plate_mode not in modes:
        eventlog.log_e_param('#/plate')
        raise JnodeError

    # unpack #/trace
    trace_mode = jnode.get('trace', LrAnnoConf.MODE_NEVER)
    if trace_mode not in modes:
        eventlog.log_e_param('#/trace')
        raise JnodeError

    # unpack #/dir
    dir_mode = jnode.get('dir', LrAnnoConf.MODE_NEVER)
    if dir_mode not in modes:
        eventlog.log_e_param('#/dir')
        raise JnodeError

    # unpack #/lane
    lane_mode = jnode.get('lane', LrAnnoConf.MODE_NEVER)
    if lane_mode not in modes:
        eventlog.log_e_param('#/lane')
        raise JnodeError

    # unpack #/center
    center_mode = jnode.get('center', LrAnnoConf.MODE_NEVER)
    if center_mode not in modes:
        eventlog.log_e_param('#/center')
        raise JnodeError

    # unpack #/vehicle
    vehicle_bbox_mode = jnode.get('vehicle-bbox', LrAnnoConf.MODE_NEVER)
    if vehicle_bbox_mode not in modes:
        eventlog.log_e_param('#/vehicle-bbox')
        raise JnodeError

    vehicle_cuboid_mode = jnode.get('vehicle-cuboid', LrAnnoConf.MODE_NEVER)
    if vehicle_cuboid_mode not in modes:
        eventlog.log_e_param('#/vehicle-cuboid')
        raise JnodeError

    return LrAnnoConf(dir_mode, lane_mode, plate_mode, trace_mode, center_mode, vehicle_bbox_mode, vehicle_cuboid_mode)


def decode__ra_anno_conf(jnode: Dict, yellow_speed: int, red_speed: int) -> RaAnnoConf:
    """
    :param jnode:
        main.conf#/Renderer/profile/vehicle/ra
    :param yellow_speed:
    :param red_speed:

    :raise JnodeError:
    """
    modes = [RaAnnoConf.MODE_NEVER, RaAnnoConf.MODE_IF_PLATE, RaAnnoConf.MODE_ALWAYS]

    # unpack #/target
    target_mode = jnode.get('target', RaAnnoConf.MODE_NEVER)
    if target_mode not in modes:
        eventlog.log_e_param('#/target')
        raise JnodeError

    # unpack #/vehicle
    vehicle_mode = jnode.get('vehicle', RaAnnoConf.MODE_NEVER)
    if vehicle_mode not in modes:
        eventlog.log_e_param('#/vehicle')
        raise JnodeError

    # unpack #/speed
    speed_mode = jnode.get('speed', RaAnnoConf.MODE_NEVER)
    if speed_mode not in modes:
        eventlog.log_e_param('#/speed')
        raise JnodeError

    # unpack #/plate
    dir_mode = jnode.get('dir', RaAnnoConf.MODE_NEVER)
    if dir_mode not in modes:
        eventlog.log_e_param('#/dir')
        raise JnodeError

    # unpack #/len
    len_mode = jnode.get('len', RaAnnoConf.MODE_NEVER)
    if len_mode not in modes:
        eventlog.log_e_param('#/len')
        raise JnodeError

    return RaAnnoConf(yellow_speed, red_speed, len_mode, dir_mode, target_mode, speed_mode, vehicle_mode,
                      radar_present=True)

def decode__markings_conf(jnode: Dict) -> MarkingsConf:
    """
    :param jnode: main.conf#/Renderer/profile/annotation/markings
    """
    lines2d_f = []
    lines3d_f = []

    color_map = {
        'red': (255, 0, 0),
        'green': (0, 255, 0),
        'blue': (0, 0, 255),
        'yellow': (255, 255, 0),
        'magenta': (255, 0, 255),
        'cyan': (0, 255, 255)
    }

    for line_f_jnode in jnode.get('line-f', []):
        active = line_f_jnode.get('active', True)
        if not active:
            continue

        color = line_f_jnode.get('color')
        rgb = color_map.get(color, (0, 255, 0))

        point3 = line_f_jnode.get('point3')
        point2 = line_f_jnode.get('point2')
        if point2 is not None:
            x0, x1, y0, y1 = point2
            dist3 = line_f_jnode['dist3']
            lines2d_f.append(MarkingLine2Conf(RefPoint2d(x0 / x1, y0 / y1), dist3, rgb))
        elif point3 is not None:
            f, l, u = point3
            dist3 = line_f_jnode['dist3']
            lines3d_f.append(MarkingLine3Conf(RefPoint3d(f, l, u), dist3, rgb))

    lines2d_l = []
    lines3d_l = []

    for line_l_jnode in jnode.get('line-l', []):
        active = line_l_jnode.get('active', True)
        if not active:
            continue

        color = line_l_jnode.get('color')
        rgb = color_map.get(color, (255, 255, 0))

        point3 = line_l_jnode.get('point3')
        point2 = line_l_jnode.get('point2')
        if point2 is not None:
            x0, x1, y0, y1 = point2
            dist3 = line_l_jnode['dist3']
            lines2d_l.append(MarkingLine2Conf(RefPoint2d(x0 / x1, y0 / y1), dist3, rgb))
        elif point3 is not None:
            f, l, u = point3
            dist3 = line_l_jnode['dist3']
            lines3d_l.append(MarkingLine3Conf(RefPoint3d(f, l, u), dist3, rgb))

    return MarkingsConf(lines2d_f, lines2d_l, lines3d_f, lines3d_l)

def decode__renderer_conf(jnode: Dict, yellow_speed: int, red_speed: int, lr_store_tout: int) -> RendererConf:
    """
    :param jnode: main.conf#/Renderer/profile/vehicle
    :param yellow_speed:
    :param red_speed:
    :param lr_store_tout:
    """
    vehilce_jnode = jnode.get('vehicle') or dict()
    lr_jnode = vehilce_jnode.get('lr')
    lr_anno = decode__lr_anno_conf(lr_jnode) if lr_jnode else LrAnnoConf()

    ra_jnode = vehilce_jnode.get('ra')
    if ra_jnode:
        ra_anno = decode__ra_anno_conf(ra_jnode, yellow_speed, red_speed)
    else:
        ra_anno = RaAnnoConf(yellow_speed, red_speed, radar_present=False)

    tl_args: List[TlConf] = decode__tl_conf(jnode)
    markings = decode__markings_conf(jnode.get('markings', {}))

    return RendererConf(lr_anno, ra_anno, tl_args, lr_store_tout, markings)
